functor

import
   Tk
   
export
   odbc: TclODBC

define

   local
      fun {ExtractRecs S}
	 fun {ExtractRec S1 ?Rest}
	    case S1 of nil then Rest = nil nil
	    [] &{|R then
	       S2 S3 S4 S5 F in
	       {String.token R &} S2 S3}
	       F = {StringToAtom S2}
	 % check end of rec
	       {String.token S3 &} S4 S5}
	       if S4 == nil then Rest = S5 [F]
	       else F|{ExtractRec S3 Rest} end
	    else
	       S2 S3 S4 F in
	       {String.token S1 32 S2 S3}
               % check end of rec
	       {String.token S2 &} S4 _}
	       if {Not S4 == S2} then
		  Rest = S3
		  [{StringToAtom S4}]
	       else
		  F = {StringToAtom S2} 
		  if F == '' then {ExtractRec S3 Rest}
		  else F|{ExtractRec S3 Rest} end
	       end
	    end
	 end
	 Rest
      in
	 case S of nil then nil
	 [] 32|&{|&{|R then
	    Rec = {ExtractRec &{|R Rest} in
	    Rec|{ExtractRecs Rest}
	 [] &{|&{|R then
	    Rec = {ExtractRec &{|R Rest} in
	    Rec|{ExtractRecs Rest}
	 [] 32|&{|R then
	    Rec = {ExtractRec R Rest} in
	    Rec|{ExtractRecs Rest}
	 [] &{|R then
	    Rec = {ExtractRec R Rest} in
	    Rec|{ExtractRecs Rest}
	 else
	    Rec = {ExtractRec S Rest} in
	    Rec|{ExtractRecs Rest}
	 end
      end

   in
      class TclODBC
	 from BaseObject

% Load the Tcl extension
	 meth init
	    {Tk.send package('require' tclodbc)}
	 end

% Unload the Tcl extension
	 meth forget
	    {Tk.send package(forget tclodbc)}
	 end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% System methods

% Create or remove a data source, see examples.  Also, create database files. 
	 meth configure(...) = M
	    Op = {CondSelect M operation nil}
	    Driver = {CondSelect M driver nil}
	    Args = {CondSelect M args nil}
	 in
	    if Args == nil then
	       raise missedArgs(M) end
	    else
	       Op1 Driver1 in
	       if Op == nil then Op1 = add_dsn else Op1 = Op end
	       if Driver == nil then Driver1 = 'Microsoft Access Driver (*.mdb)' else Driver1 = Driver end
	       {Tk.send database(configure Op1 Driver1 Args)}
	    end
	 end

% Return all ODBC drivers availble in the system.
	 meth drivers(?Ds)
	    Ds = {ExtractRecs {Tk.return database(drivers)}}
	 end

% Return all data sources availble in the system.
	 meth datasources(?DSs)
	    DSs = {ExtractRecs {Tk.return database(datasources)}}
	 end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Database

% Open a data source
% Or
% Connect directly to database file using connection string
	 meth connect(...) = M
	    DS = {CondSelect M name nil} in
	    if DS == nil then raise missedDataSourceName(M) end
	    else
	       STR = {CondSelect M str nil} in
	       if STR == nil then
		  User = {CondSelect M user nil} in
		  if User == nil then {Tk.send database(connect DS DS '' '')}
		  else
		     Pwd = {CondSelect M password ''} in
		     {Tk.send database(connect DS DS User Pwd)}
		  end
	       else
		  {Tk.send database(connect DS STR)}
	       end
	    end
	 end
	 
% Close a data source
	 meth disconnect(DS)
	    {Tk.send DS(disconnect)}
	 end

% Set ODBC option
	 meth set(DS Option Value)
	    {Tk.send DS(set Option Value)}
	 end
	 
% Get ODBC option value
	 meth get(DS Option ?Value)
	    Value = {Tk.returnAtom DS(get Option)}
	 end
	 
% Set and unset the autocommit option
	 meth autocommit(DB OnOff)
	    {Tk.send DB(set autocommit OnOff)}
	 end
   
% Rollback a previous action when autocommit have been set off
	 meth rollback(DB)
	    {Tk.send DB(rollback)}
	 end

% Commit a previous action when autocommit have been set off
	 meth commit(DB)
	    {Tk.send DB(commit)}
	 end

% Return all the table names of a database
% [[TABLE_QUALIFIER TABLE_OWNER TABLE_NAME TABLE_TYPE REMARKS] ... ] 
% TABLE_TYPE either consists of the atom(s): '{SYSTEM' 'TABLE}' or 'TABLE'
	 meth tables(DB Ts)
	    Ts = {ExtractRecs {Tk.return DB(tables)}}
	 end
   
% Return all the columns description of a database in a continuous list of the form:
%[[TABLE_QUALIFIER TABLE_OWNER TABLE_NAME COLUMN_NAME DATA_TYPE TYPE_NAME PRECISION LENGTH SCALE RADIX NULLABLE REMARKS] ... ]
% i.e., each column in the database/table is defined by 12 items in the returned list of lists.
	 meth dbcolumns(DB Cs)
	    Cs = {ExtractRecs {Tk.return DB(columns)}}
	 end
   
% Return all the column names of a database table
	 meth tablecolumns(DB Table Cs)
	    Cs = {ExtractRecs {Tk.return DB(columns Table)}}
	 end
   
% Return all the index column names of a database table
	 meth indexes(DB Table Is)
	    Is = {ExtractRecs {Tk.return DB(indexes Table)}}
	 end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SQL

% Exec the sql 
	 meth sql(DB SQL)
	    {Tk.send DB(SQL)}
	 end
   
% Return the sql result
	 meth sqlReturn(DB SQL Rs)
	    Rs = {ExtractRecs {Tk.return DB(q(SQL))}}
	 end

% Statement are used with large tables when the table cannot be loaded in memory
% Define a given sql statement
% statement id <sql clause>|tables|columns|typeinfo ?argtypedefs?
% Examples:
% statement(db s "select fullname from article where id"#"="#"132") 
% statement(db s2 "select fullname from article where id=?"#" INTEGER"
	 meth statement(DB Stmt SQL)
	    {Tk.send DB(statement Stmt SQL)}
	 end

% Execute the sql statement
% Using the previous examples:
% stmt(s Rs)
% stmt(s(132) Rs)
	 meth stmt(Stmt Rs)
	    Rs = {ExtractRecs {Tk.return Stmt}}
	 end

	 meth stmt_extended(Stmt TypeNames Values Rs)
	    Rs = {ExtractRecs {Tk.return Stmt(q(TypeNames) q(Values))}}
	 end

% Execute statement
	 meth execute(Stmt)
	    {Tk.send Stmt(execute)}
	 end

% The call to execute should be followed by a call to fetch
	 meth fetch(Stmt Rs)
	    Rs = {ExtractRecs {Tk.return Stmt(fetch)}}
	 end

% Delete a statement
	 meth drop(Stmt)
	    {Tk.send Stmt(drop)}
	 end

% Returns the cloumns 'label' of the table of the SQL statement
	 meth cols(Stmt Rs)
	    Rs = {ExtractRecs {Tk.return Stmt(columns)}}
	 end

% returns the colums ATTRIBUTES of the table of the SQL statement
% ATTRIBUTES can be a cobination of:
% [label name displaysize type typename precision scale nullable updatable tablename qualifiername owner]
	 meth cols_extended(Stmt Atts Rs)
	    Rs = {ExtractRecs {Tk.return Stmt(columns b(Atts))}}
	 end

      end
   end
end


